/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is 
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
GOOD { clip: auto; }
GOOD { clip: rect( auto auto auto auto ) }
GOOD { clip: rect(auto auto auto auto) }
GOOD { clip: rect( 12pt auto auto auto ) }

GOOD { clip: rect( auto 12pt auto auto ) }
GOOD { clip: rect( auto auto 12pt auto ) }
GOOD { clip: rect( auto auto auto 12pt ) }
GOOD { clip: rect( 12pt 13em auto auto ) }

GOOD { clip: rect( 12pt 13em 14pc auto ) }
GOOD { clip: rect( 12pt 13em 14pc 15in ) }
GOOD { clip: rect ( 12pt 13em 14pc 15in ) }

GOOD { clip: rect( auto, auto, auto, auto ) }
GOOD { clip: rect(auto  , auto  , auto  ,   auto) }
GOOD { clip: rect( 12pt, auto, auto, auto, ) }

GOOD { clip: rect( auto, 12pt, auto, auto ) }
GOOD { clip: rect( auto, auto, 12pt, auto ) }
GOOD { clip: rect( auto, auto, auto, 12pt ) }
GOOD { clip: rect( 12pt, 13em, auto, auto ) }

GOOD { clip: rect( 12pt, 13em, 14pc, auto ) }
GOOD { clip: rect( 12pt, 13em, 14pc, 15in ) }
GOOD { clip: rect ( 12pt, 13em, 14pc, 15in ) }

BAD { clip: rect( 1pt ) }
BAD { clip: rect( 1pt 2pt ) }
BAD { clip: rect( 1pt 2pt 3pt ) }
BAD { clip: rect( 1 2 3 4 ) }
BAD { clip: rect( auto ) }
BAD { clip: rect( auto auto) }
BAD { clip: rect( auto auto auto) }
BAD { clip: rect( auto auto auto) }
BAD { clip: rect( auto, auto, auto, auto, ) }
BAD { clip: rect ( 12pt, 13em, 14pc, 15in, ) }
