/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is 
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

@namespace url(http://www.w3.org/1999/xhtml); /* set default namespace to HTML */

select[size] 
  {
    margin                : 0px;
  }
  
.select-container-treerows 
  {
    border-left           : 1px solid threeddarkshadow;
    border-top            : 1px solid threeddarkshadow;
    border-right          : 1px solid threedface;
    border-bottom         : 1px solid threedface;
  } 
  
option
  {
    font                  : -moz-list;  
    border                : 1px solid transparent;
    color                 : windowtext;
  }  

select[size] option[selected]
  {
    background-color      : highlight;
    color                 : highlighttext;
  }  

select option[selected],
select[size="1"] option[selected]
  {
    background-color      : transparent;
    color                 : windowtext;
  } 
  
/* menulist based select */

select, select[size="1"]
  {
    border-left           : 1px solid threedshadow;
    border-top            : 1px solid threedshadow;
    border-right          : 1px solid threedhighlight;
    border-bottom         : 1px solid threedhighlight;
    background-color      : -moz-field;
    color                 : menutext;
    font                  : -moz-list;
    margin                : 1px 5px 2px 5px;
  }

select[disabled]
  {
    background-color      : threedface;
  }
    
/* icon display frame */
.select-menulist-internal-box
  {
    border-left           : 1px solid threeddarkshadow;
    border-top            : 1px solid threeddarkshadow;
    border-right          : 1px solid threedface;
    border-bottom         : 1px solid threedface;
    -moz-user-focus       : ignore;
  }  

.select-menulist-display-box,
select[open="true"]:focus > .select-menulist-internal-box > .select-menulist-display-box
  {
    border                : 1px solid transparent;
    margin                : 1px;
    background-color      : -moz-field;
    color                 : menutext;
  }     
    
select:focus > .select-menulist-internal-box > .select-menulist-display-box
  {
    border                : 1px dotted #F5DB95;
    background-color      : highlight;
    color                 : highlighttext;
  }
      
/* text display frame */
.select-menulist-text 
  {
    padding               : 0px 2px 0px 1px;
    margin                : 0px;
  }
    
/* drop marker display frame */
.select-menulist-dropmarker-box-1
  {
    border-top            : 1px solid threedface;
    border-left           : 1px solid threedface;
    border-bottom         : 1px solid threeddarkshadow;
    border-right          : 1px solid threeddarkshadow;
    background-color      : threedface;
  }
  
.select-menulist-dropmarker-box-2
  {
    border-top            : 1px solid threedhighlight;
    border-left           : 1px solid threedhighlight;
    border-bottom         : 1px solid threedshadow;
    border-right          : 1px solid threedshadow;
    padding               : 2px 1px 2px 1px;
  }

.select-menulist-dropmarker
  {
    list-style-image      : url("chrome://global/skin/scroll-down.gif");
  }
  
.select-menulist-dropmarker[disabled="true"]
  {
    list-style-image      : url("chrome://global/skin/scroll-down-disabled.gif");
    padding               : 2px;
  }

select:hover:active > .select-menulist-internal-box > .select-menulist-dropmarker-box-1[open="true"]
  { 
    border                : 1px solid threedshadow;
  }
    
select:hover:active > .select-menulist-internal-box > .select-menulist-dropmarker-box-1 > .select-menulist-dropmarker-box-2[open="true"]
  {
    border                : 1px solid threedface;
    padding               : 3px 0px 1px 2px;
  }

/* rules for popups and separators associated with menulists */
select > .select-menulist-menupopup 
  {
    background-color      : -moz-field;
    border                : 1px solid #000000;
    min-width             : 0px;
  }

select > option 
  {
    padding               : 0px 7px !important;
    min-width             : 0px;  /* should be in content*/
    max-width             : none; /* should be in content*/
    border                : 1px solid transparent;
  }

select:focus > option[menuactive="true"]
  {
    border                : 1px dotted #F5DB95;
    background-color      : highlight;
    color                 : highlighttext;
  }


