/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is 
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

@namespace url(http://www.w3.org/1999/xhtml); /* set default namespace to HTML */

/* stylesheet for XUL <checkbox> element */

/* default checkbox for dialogs */
/* outer frame */

  input[type="checkbox"]
    {
      border              : 1px solid transparent;
      padding             : 2px !important;
      margin              : 0px;
      cursor              : default;
    }

  .input-checkbox-checkmark-box-1
    {
      border-top          : 1px solid threedshadow;
      border-left         : 1px solid threedshadow;
      border-right        : 1px solid threedhighlight;
      border-bottom       : 1px solid threedhighlight;
      background-color    : window;
      width               : 13px;
      height              : 13px;
      -moz-user-focus     : ignore;
    }

  .input-checkbox-checkmark-box-1[disabled]
    {
      background-color    : threedface;
    }

  .input-checkbox-checkmark-box-2
    {
      border-left         : 1px solid threeddarkshadow;
      border-top          : 1px solid threeddarkshadow;
      border-right        : 1px solid threedface;
      border-bottom       : 1px solid threedface;
      padding             : 1px;
      width               : 11px;
      height              : 11px;
    }  

  /* ensure that no list-style-image is inherited in from the Outside */
  .input-checkbox-check
    {
      list-style-image    : none;
    }

  .input-checkbox-check[checked]
    {
      list-style-image    : url(chrome://global/skin/check-check.gif);
    }
  
  .input-checkbox-check[checked][disabled]
    {
      list-style-image    : url(chrome://global/skin/check-check-disabled.gif);
    }

  input[type="checkbox"]:hover:active > .input-checkbox-checkmark-box-1
    {
      background-color    : threedface;
    }

/* text/icon frames */
  input[type="checkbox"]:focus
    { 
      border              : 1px dotted threeddarkshadow;
      padding             : 0px;
    }
    
  .input-checkbox-checkmark-box-1
    {
      -moz-user-focus     : none; 
    }

    
