/** -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 * 
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation. Portions created by Netscape are
 * Copyright (C) 2001 Netscape Communications Corporation. All
 * Rights Reserved.
 * 
 * Contributor(s): 
 *   Stuart Parmenter <pavlov@netscape.com>
 */

#include "nsISupports.idl"
#include "gfx2types.idl"

native HDC(HDC);
native HBITMAP(HBITMAP);
native LPBITMAPINFOHEADER(LPBITMAPINFOHEADER);

%{C++
#include <windows.h>
%}

/**
 * gfxIImageFrameWin interface
 *
 * @author Stuart Parmenter <pavlov@netscape.com>
 * @version 0.0
 */
[uuid(2206658e-1dd2-11b2-82ae-e853b4c0a9d5)]
interface gfxIImageFrameWin : nsISupports
{
  void getDIB(out HBITMAP aBitmap, out HBITMAP aMask);
  void getDDB(out HBITMAP aBitmap, out HBITMAP aMask);
  void getAlphaStuff(out HBITMAP aBitmap, out LPBITMAPINFOHEADER aInfoHeader, out voidPtr aBits);
  void getBitmap(in HDC aHDC, out HBITMAP aBitmap, out HBITMAP aMask, out gfx_depth aAlphaDepth);
  readonly attribute gfx_depth alphaDepth;
};
